// Copyright 2006-2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-engine
// File: ocr-recognize-line.h
// Purpose: ILineOCR interface
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifndef h_ocr_recognize_line_
#define h_ocr_recognize_line_

#include "colib.h"

namespace ocropus {


    /// A line recognizer that wraps around IRecognizeLine, ISegmentLine
    /// and ITrainChars to provide support for automatic line rescaling
    /// and retraining. After the recognition, you can give the result to this
    /// object and get some data for retraining.
    ///
    /// Note that ILineOCR does not contain any language model.
    /// That's why the recognition is finished outside of an ILineOCR,
    /// and after it finished you can return to ILineOCR for retraining.
    struct ILineOCR {
        // _______________________    bookkeeping    ___________________________
        
        virtual void inject(colib::ITrainChars    *) = 0;
        virtual void inject(colib::IRecognizeLine *) = 0;
        virtual void inject(colib::ISegmentLine   *) = 0;
        
        virtual colib::ITrainChars    *take_TrainChars()    = 0;
        virtual colib::IRecognizeLine *take_RecognizeLine() = 0;
        virtual colib::ISegmentLine   *take_SegmentLine()   = 0;

        /// Give a description of the line (for debugging purposes).
        /// (the default implementation is provided for some tools I could name
        ///  that don't handle varargs nicely)
        virtual void setDescription(const char *format, ...) {}

        // ________________________   recognition   ____________________________
        
        /// Give a grayscale image with a mask to the recognizer.
        virtual void set(colib::bytearray &gray,
                         colib::bytearray &mask) = 0;

        /// Get the oversegmentation of the original image.
        /// Might have gaps.
        virtual void getOversegmentation(colib::intarray &) = 0;
         
        /// Recognize and fill the char lattice.
        /// Note that idmap stuff is by design encapsulated inside ILineOCR. 
        virtual void recognize(colib::ICharLattice &) = 0;
        
        // ________________________   adaptation   _____________________________

        /// Given a result of the best path search,
        /// train the given ITrainChars instance.
        /// Note that it's better to train with adapt() then
        /// through getSegmentation() since adapt() might use
        /// rescaled or otherwise transformed characters
        /// better suited for adaptation.
        virtual void adapt(colib::nustring &ocr_text,
                           colib::floatarray &costs,
                           colib::intarray &ids,
                           float adaption_threshold) = 0;

        /// Return "true" segmentation of the original image.
        /// Might have gaps.
        virtual void getSegmentation(colib::intarray &) = 0;

        virtual ~ILineOCR() {}
    };
    
    //ILineOCR *make_LineOCR();
    
    /// A minimal LineOCR implementation that doesn't support rescaling.
    ILineOCR *make_SimpleLineOCR();
}

#endif
