// Copyright 2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: 
// File: ocr-recognize-page-new.cc
// Purpose: 
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de, www.ocropus.org

#include "ocr-recognize-page.h"
#include "sysutil.h"
#include "tesseract.h"

using namespace colib;

namespace {
    rectangle unite_rectangles(narray<rectangle> &rects) {
        rectangle result;
        for(int i = 0; i < rects.length(); i++) {
            result.include(rects[i]);
        }
        return result;
    }
}

namespace ocropus {
     
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   ILineOCR &ocr,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   float adaption_threshold,
                   int pageno) {
        if(pageno >= 0) lines.processPage(pageno);
        double start = now();
        
        result.setLinesCount(lines.linesCount());
        result.setWidth(lines.pageWidth());
        result.setHeight(lines.pageHeight());
        result.setDescription(lines.pageDescription());

        for(int i = 0; i < lines.linesCount(); i++) {
            bytearray gray, mask;
            lines.line(gray, mask, i);
            ocr.set(gray, mask);
            ocr.recognize(lattice);

            // here the magical connection between lattice and bestpath works...

            nustring text;
            floatarray costs;
            intarray ids;
            intarray states;
            bestpath.bestpath(text, costs, ids, states);
            ocr.adapt(text, costs, ids, adaption_threshold);

            result.setText (text, i);
            result.setCosts(costs, i);
            result.setBbox(lines.bbox(i), i);
        }
        double LA_time = lines.getCurrentPageElapsedTime();
        double end = now();
        char buf[1000];
        sprintf(buf, "time elapsed:\n"
                     "  layout analysis: %.2f sec,\n"
                     "  recognition: %.2f sec,\n"
                     "Total: %.2f sec\n",
                     LA_time, end - start, LA_time + end - start);
        result.setTimeReport(buf);
    }

    /// 1-pass recognition.
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &rec,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   float adaption_threshold,
                   int pageno) {
        autodel<ILineOCR> ocr(make_SimpleLineOCR());
        ocr->inject(&lineseg);
        ocr->inject(&rec);
        recognize(result, lines, *ocr, lattice, bestpath,
                  adaption_threshold, pageno);
        ocr->take_RecognizeLine();
        ocr->take_SegmentLine();
    }

    /// 1-pass recognition with training a ITrainChars.
    void recognize(RecognizedPage &result,
                   ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &rec,
                   colib::ICharLattice &lattice,
                   colib::IBestPath &bestpath,
                   ITrainChars &trainable,
                   float adaption_threshold,
                   int pageno) {
        autodel<ILineOCR> ocr(make_SimpleLineOCR());
        ocr->inject(&lineseg);
        ocr->inject(&rec);
        ocr->inject(&trainable);
        recognize(result, lines, *ocr, lattice, bestpath,
                  adaption_threshold, pageno);
        ocr->take_RecognizeLine();
        ocr->take_SegmentLine();
        ocr->take_TrainChars();
    }
    
    
    /// 2-pass recognition.
    void recognize(RecognizedPage &result,
                   ocropus::ILines &lines,
                   colib::ISegmentLine &lineseg,
                   colib::IRecognizeLine &pass1,
                   colib::ICharLattice &lattice1,
                   colib::IBestPath &bestpath1,
                   colib::ITrainChars &trainable,
                   colib::IRecognizeLine &pass2,
                   colib::ICharLattice &lattice2,
                   colib::IBestPath &bestpath2,
                   float adaption_threshold,
                   int pageno) {
        RecognizedPage temp;
        if(pageno >= 0) lines.processPage(pageno);
        double start = now();
        recognize(temp,   lines, lineseg, pass1, lattice1, bestpath1,
                  trainable, adaption_threshold, -1);
        double between = now();
        recognize(result, lines, lineseg, pass2, lattice2, bestpath2,
                  adaption_threshold, -1);
        double end = now();

        double LA_time = lines.getCurrentPageElapsedTime();
        double p1_time = between - start;
        double p2_time = end - between;

        char buf[1000];
        sprintf(buf, "time elapsed:\n"
                     "  layout analysis: %.2f sec\n"
                     "  first pass: %.2f sec\n"
                     "  second pass: %.2f sec\n"
                     "Total: %.2f sec\n",
                     LA_time, p1_time, p2_time, LA_time + p1_time + p2_time);
        result.setTimeReport(buf);
    }
    
    void tesseract_recognize_blockwise(RecognizedPage &result, colib::bytearray &gray, colib::intarray &pageseg) {
        //double start = now();
        narray<nustring> text;
        narray<narray<rectangle> > bboxes;
        narray<floatarray> costs;
        tesseract_recognize_blockwise_and_split_to_lines(text, bboxes, costs, gray, pageseg);
        result.setWidth(gray.dim(0));
        result.setHeight(gray.dim(1));
        result.setLinesCount(text.length());
        for(int i = 0; i < text.length(); i++) {
            result.setText(text[i], i);
            result.setBbox(unite_rectangles(bboxes[i]), i);
            result.setCosts(costs[i], i);
        }
        // description
        // Time report        
    }

    void tesseract_recognize_blockwise(RecognizedPage &result,
                                       ILines &lines,
                                       int pageno) {
        if(pageno >= 0) lines.processPage(pageno);

        double start = now();
        tesseract_recognize_blockwise(result, lines.grayPage(),
                                              lines.segmentation());
        result.setDescription(lines.pageDescription());
        
        double LA_time = lines.getCurrentPageElapsedTime();
        double end = now();
        char buf[1000];
        sprintf(buf, "time elapsed:\n"
                     "  layout analysis: %.2f sec\n"
                     "  recognition: %.2f sec\n"
                     "Total: %.2f sec\n",
                     LA_time, end - start, end - start + LA_time);
        result.setTimeReport(buf);
    }

}

