// -*- C++ -*-

// Copyright 2006-2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-utils
// File: narray-io.h
// Purpose: I/O utilities
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifndef h_narray_io_
#define h_narray_io_

#include <stdio.h>
#include <stdlib.h>
#include "colib.h"

namespace ocropus {

// ____________________   printing narrays for humans   _______________________

/// Print a 2D array formatted with to the given printf-format string.
    template<class T>
        void printf_matrix(const char *format, colib::narray<T> &a) {
        ASSERT(a.rank() == 2);
        for(int i = 0; i < a.dim(0); i++) {
            for(int j = 0; j < a.dim(1); j++) {
                if(j)
                    putchar(' ');
                printf(format, a(i,j));
            }
            putchar('\n');
        }
    }

/// Print any array as 1D vector formatted with the given printf-format string.
    template<class T>
        void printf_vector(const char *format, colib::narray<T> &a) {
        for(int i = 0; i < a.length1d(); i++) {
            if(i)
                putchar(' ');
            printf(format, a.at1d(i));
        }
        putchar('\n');
    }

    template<class T>
        void printf_narray(const char *format, colib::narray<T> &a) {
        if(a.rank() == 2)
            printf_matrix(format, a);
        else
            printf_vector(format, a);
    }

    inline void print(colib::doublearray &a) {printf_narray("%g", a);}
    inline void print(colib::floatarray &a) {printf_narray("%g", a);}
    inline void print(colib::intarray &a) {printf_narray("%d", a);}


// ________________________   text I/O for narrays  ___________________________

    void text_write_1d(FILE *, colib::intarray &);
    void text_read_1d(FILE *, colib::intarray &);
    void text_write(FILE *, colib::floatarray &);
    inline void text_write(colib::floatarray &a, FILE *f) {text_write(f, a);}
    void text_read(FILE *, colib::floatarray &);
    inline void text_read(colib::floatarray &a, FILE *f) {text_read(f, a);}
    void text_write(FILE *, colib::doublearray &);
    inline void text_write(colib::doublearray &a, FILE *f) {text_write(f, a);}
    void text_read(FILE *, colib::doublearray &);
    inline void text_read(colib::doublearray &a, FILE *f) {text_read(f, a);}
    void text_write(FILE *, colib::intarray &);
    inline void text_write(colib::intarray &a, FILE *f) {text_write(f, a);}
    void text_read(FILE *, colib::intarray &);
    inline void text_read(colib::intarray &a, FILE *f) {text_read(f, a);}
    void text_write(FILE *, colib::bytearray &);
    inline void text_write(colib::bytearray &a, FILE *f) {text_write(f, a);}
    void text_read(FILE *, colib::bytearray &);
    inline void text_read(colib::bytearray &a, FILE *f) {text_read(f, a);}

// ______________________   binary I/O for primitives   _______________________

    inline void write_int32(int32_t n, FILE *f) {
        fputc(n >> 24, f);
        fputc(n >> 16, f);
        fputc(n >>  8, f);
        fputc(n,       f);
    }
    inline void write_int32(FILE *f, int32_t n) {write_int32(n,f);}
    inline int32_t read_int32(FILE *f) {
        int i = fgetc(f);
        i = (i << 8) | fgetc(f);
        i = (i << 8) | fgetc(f);
        i = (i << 8) | fgetc(f);
        return i;
    }

    inline void bin_read(FILE *f, colib::rectangle &r) {
        r.x0 = read_int32(f);
        r.y0 = read_int32(f);
        r.x1 = read_int32(f);
        r.y1 = read_int32(f);
    }

    inline void bin_read(colib::rectangle &r, FILE *f) {bin_read(f,r);}

    inline void bin_write(FILE *f, colib::rectangle &r) {
        write_int32(f, r.x0);
        write_int32(f, r.y0);
        write_int32(f, r.x1);
        write_int32(f, r.y1);
    }

    inline void bin_write(colib::rectangle &r, FILE *f) {bin_write(f,r);}


    inline int read_raw(FILE *f, colib::bytearray &a) {
        return fread(a.data, 1, a.length1d(), f);
    }
    inline int read_raw(colib::bytearray &a, FILE *f) {return read_raw(f, a);}
    inline int write_raw(FILE *f, colib::bytearray &a) {
        return fwrite(a.data, 1, a.length1d(), f);
    }
    inline int write_raw(colib::bytearray &a, FILE *f) {return write_raw(f, a);}

// ______________________________   misc   ____________________________________

    void parse_vector(colib::doublearray &,const char *);
    inline void parse_vector(colib::intarray &result, const char *s) {
        colib::doublearray tmp;
        parse_vector(tmp, s);
        colib::copy(result, tmp);
    }

    void bin_write_1d(FILE *, colib::intarray &);
    void bin_read_1d(FILE *, colib::intarray &);
    void bin_write_mnist_dims(FILE *, colib::intarray &);
    void bin_read_mnist_dims(FILE *, colib::intarray &);
    void bin_write(FILE *, colib::bytearray &);
    inline void bin_write(colib::bytearray &a, FILE *f) {bin_write(f, a);}
    void bin_read(FILE *, colib::bytearray &);
    inline void bin_read(colib::bytearray &a, FILE *f) {bin_read(f, a);}
    void bin_append(FILE *, colib::bytearray &);
    inline void bin_append(colib::bytearray &a, FILE *f) {bin_append(f, a);}

    inline void bin_read(const char *path, colib::bytearray &a) {
        bin_read(colib::stdio(path, "rb"), a);
    }
    inline void bin_read(colib::bytearray &a, const char *path) {
        bin_read(path, a);
    }
    inline void bin_write(const char *path, colib::bytearray &a) {
        bin_write(colib::stdio(path, "wb"), a);
    }
    inline void bin_write(colib::bytearray &a, const char *path) {
        bin_write(path, a);
    }
    inline void bin_append(const char *path, colib::bytearray &a) {
        FILE *test = fopen(path, "rb");
        fclose(test);
        if(test) {
            colib::stdio f(path, "r+b");
            bin_append(f, a);
        } else {
            colib::stdio f(path, "wb");
            bin_write(f, a);
        }
    }
    inline void bin_append(colib::bytearray &a, const char *path) {
        bin_append(path, a);
    }

    void bin_write_nustring(FILE *f, colib::nustring &a);
    void bin_read_nustring(FILE *f, colib::nustring &a);

};

#endif
