// Copyright 2006-2007 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocropus
// File: ocropus.cc
// Purpose: the main program file for OCRopus
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de, www.ocropus.org

#include "colib.h"
#include "imgio.h"
#include "imglib.h"
#include "triv-hocr.h"
#include "ocr-layout-rast.h"
#include "ocr-deskew-rast.h"
#include "ocr-doc-clean.h"
#include "ocr-text-image-seg.h"
#include "ocr-recognize-line.h"
#include "ocr-recognize-page.h"
#include "ocr-binarize-sauvola.h"
#include "langmod-aspell.h"
#include "langmod-shortest-path.h"
#include "segmentation.h"
#include "tesseract.h"
#include "ocr-utils.h"
#include "ocr-segmentations.h"
#include "version.h"
#include "logging.h"
#include "bpnet.h"
#include "ocrcomponents.h"
#include "layout-rules.h"

#define TRAINING_THRESHOLD 30

using namespace ocropus;
using namespace imgio;
using namespace imglib;
using namespace colib;


void print_help(const char* command) {
    puts("\nOcropus - open source document analysis and OCR system");
    //printf("%s\n", get_version_string());
    puts("Usage:");
    printf("    %s <image> [<image>*]  - recognize given page(s)\n", command);
    //printf("    %s -                   - read filename(s) from stdin\n", command);
    printf("    %s --help              - print this help message\n", command);
    puts("");
    puts("Options are passed through the environment only.");
    puts("The most common options are listed below (first value is default):");
    puts("    blockwise=1/0        - uses Tesseract block by block");
    puts("    quiet=0/1            - disables output of version information");
    puts("    hocr=1/0             - enables hOCR or plain text output");
    puts("    timing=0/1           - enables output of timing information");
    puts("    cleanup=0/1          - enables document image cleanup");
    puts("    tiseg=0/1            - enables text image segmentaion");
    puts("    deskew=0/1           - enables document image deskewing");
    puts("    debug_cleanup=<png>  - output cleaned document image as png");
    puts("    debug_tiseg=<png>    - output result of text image segmentation as png");
    puts("    debug_deskew=<png>   - output deskewed document image as png");
    puts("");
    puts("Refer to these websites for more information:");
    puts("    www.ocropus.org                  - Ocropus Home");
    puts("    groups.google.com/group/ocropus  - Ocropus Google Group");
    puts("");
}

namespace ocropus {
    param_bool quiet("quiet", false, "disables output of version information");
    param_bool timing("timing", false, "enables output of timing information");
    param_bool doccleanup("cleanup", false, "enables document image cleanup");
    param_bool tiseg("tiseg", false, "enables text image segmentaion");
    param_bool deskew("deskew", false, "enables document image deskewing");
    param_bool hocr("hocr", true, "output hOCR");
    param_bool blockwise("blockwise", true, "Tesseract blockwise mode");
    param_bool spellcheck("spell", true,
            "applies spellchecking at the first pass");
    param_bool nocatch("nocatch", false, "don't catch exceptions");
    param_bool recolor("recolor", false,
            "recolor segmentation outputs (for easier display)");
    param_string binarize_name("binarize", "sauvola", "binarizer");
    param_string segment_page_name("segment_page", "rast", "page segmenter");
    param_string segment_line_name("segment_line", "curved", "line segmenter");
    param_string language_model_name("language_model", "aspell",
            "language_model");
    param_string format("format", NULL, "input file format");
}
;

IBinarize *make_Binarize(const char *name) {
    if (!strcmp(name, "sauvola"))
        return make_BinarizeBySauvola();
    if (!strcmp(name, "range"))
        return make_BinarizeByRange();
    throw "unknown binarizer";
}

ISegmentPage *make_SegmentPage(const char *name) {
    if (!strcmp(name, "smear"))
        return make_SegmentPageBySmear();
    if (!strcmp(name, "1cp"))
        return make_SegmentPageBy1CP();
    if (!strcmp(name, "rast"))
        return make_SegmentPageByRAST();
    throw "unknown page segmenter";
}

ISegmentLine *make_SegmentLine(const char *name) {
    if (!strcmp(name, "proj"))
        return make_SegmentLineByProjection();
    if (!strcmp(name, "ccs"))
        return make_SegmentLineByCCS();
    if (!strcmp(name, "ccs1"))
        return make_ConnectedComponentSegmenter();
    if (!strcmp(name, "curved"))
        return make_CurvedCutSegmenter();
    throw "unknown line segmenter";
}

ILines *make_Lines_with_stuff(const char *input_spec) {
    ILines *lines = make_Lines(input_spec);
    lines->setPageSegmenter(make_SegmentPage(segment_page_name));
    if(deskew)
        lines->setDeskewer(make_DeskewGrayPageByRAST());
    if(doccleanup)
        lines->addCleanupBinary(make_DocClean());
    if(tiseg)
        lines->addCleanupBinary(make_TextImageSeg());
    lines->setBinarizer(make_Binarize(binarize_name));
    return lines;
}

/**
 * Checks whether a file exists and is readable.
 * 
 * @param file the name of the file to check
 * @return true if file is readable, false otherwise
 */
bool file_is_readable(const char* file) {
    FILE* fp = fopen(file, "r");
    if(fp) {
        fclose(fp);
        return true;
    }
    return false;
}

void ocropus_recognize(const char *input_spec) {
    autodel<ILines> lines(make_Lines_with_stuff(input_spec));
    autodel<ISegmentLine> segment_line;
    autodel<IRecognizeLine> recognize_line;
    autodel<ITrainChars> train_chars;
    autodel<ISearchableCharLattice> langmod1;
    autodel<ISearchableCharLattice> langmod2;
    autodel<IBestPath> bestpath1;
    
    if(!blockwise) {
        segment_line = make_SegmentLine(segment_line_name);
        recognize_line = make_TesseractRecognizeLine();
        train_chars = make_TesseractTrainChars();
        langmod1 = make_ShortestPathCharLattice();
        langmod2 = make_ShortestPathCharLattice();
        bestpath1 = make_DictBestPath(&*langmod1, make_Aspell());
    }
    if(hocr) output_hOCR_header();
    for(int page = 0; page < lines->pagesCount(); page++) {
        RecognizedPage recognized_page;
        if(blockwise) {
            tesseract_recognize_blockwise(recognized_page, *lines, page);
        } else {
            recognize(recognized_page, *lines,
                  *segment_line, *recognize_line,
                  *langmod1, *bestpath1,
                  *train_chars,
                  *recognize_line,
                  *langmod2, *langmod2,
                  50, // training threshold
                  page);
        }
        detect_headlines (recognized_page, *lines);
        detect_paragraphs(recognized_page, *lines);
        
        if(hocr) {
            output_hOCR(recognized_page);
        } else {
            for(int i = 0; i < recognized_page.linesCount(); i++) {
                nustring nus;
                recognized_page.text(nus, i);
                char *s = nus.newUtf8Encode();
                puts(s);
                delete[] s;
            }
        }
    }
    if(!blockwise) {
        langmod1.move(); // release ownership -- owned by bestpath1
    }
    if(hocr) output_hOCR_footer();
}

int real_main(int argc, char *argv[]) {
    // -- print info if not quiet --
    if (!quiet)
        fprintf(stderr, "Ocropus Alpha (%s, %s, %s, %s)\n%s\n",
                (const char*)binarize_name,
                (const char*)segment_page_name,
                (const char*)segment_line_name,
                "tesseract",
                get_version_string());

    for(int i = 1; i < argc; i++) {
        if(file_is_readable(argv[i]))
            ocropus_recognize(argv[i]);
        else {
            fprintf(stderr, "\nFile is not valid: %s\n", argv[i]);
        }
    }

    return 0;
}

int main(int argc, char **argv) {
    set_version_string(hardcoded_version_string());

    // -- check if user wants help -- 
    if (argc < 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "-h")) {
        print_help(argv[0]);
        return 0;
    }

    // -- run actual main --
    if (nocatch) {
        return real_main(argc, argv);
    } else {
        try {
            return real_main(argc,argv);
        }
        catch(const char *s) {
            fprintf(stderr, "%s: %s\n", argv[0], s);
            exit(251);
        }
    }
}
